from __future__ import absolute_import, unicode_literals

__all__ = ['AQL', 'AQLQueryCache']

from .api import APIWrapper
from .cursor import Cursor
from .exceptions import (
    AQLQueryExplainError,
    AQLQueryValidateError,
    AQLQueryExecuteError,
    AQLQueryListError,
    AQLQueryClearError,
    AQLQueryTrackingGetError,
    AQLQueryKillError,
    AQLQueryTrackingSetError,
    AQLFunctionCreateError,
    AQLFunctionDeleteError,
    AQLFunctionListError,
    AQLCacheEntriesError,
    AQLCacheClearError,
    AQLCacheConfigureError,
    AQLCachePropertiesError
)
from .formatter import (
    format_aql_cache,
    format_aql_query,
    format_aql_tracking
)
from .request import Request


class AQL(APIWrapper):
    """AQL (ArangoDB Query Language) API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    """

    def __init__(self, connection, executor):
        super(AQL, self).__init__(connection, executor)

    def __repr__(self):
        return '<AQL in {}>'.format(self._conn.db_name)

    @property
    def cache(self):
        """Return the query cache API wrapper.

        :return: Query cache API wrapper.
        :rtype: arango.aql.AQLQueryCache
        """
        return AQLQueryCache(self._conn, self._executor)

    def explain(self, query, all_plans=False, max_plans=None, opt_rules=None):
        """Inspect the query and return its metadata without executing it.

        :param query: Query to inspect.
        :type query: str | unicode
        :param all_plans: If set to True, all possible execution plans are
            returned in the result. If set to False, only the optimal plan
            is returned.
        :type all_plans: bool
        :param max_plans: Total number of plans generated by the optimizer.
        :type max_plans: int
        :param opt_rules: List of optimizer rules.
        :type opt_rules: list
        :return: Execution plan, or plans if **all_plans** was set to True.
        :rtype: dict | list
        :raise arango.exceptions.AQLQueryExplainError: If explain fails.
        """
        options = {'allPlans': all_plans}
        if max_plans is not None:
            options['maxNumberOfPlans'] = max_plans
        if opt_rules is not None:
            options['optimizer'] = {'rules': opt_rules}

        request = Request(
            method='post',
            endpoint='/_api/explain',
            data={'query': query, 'options': options}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryExplainError(resp, request)
            if 'plan' in resp.body:
                return resp.body['plan']
            else:
                return resp.body['plans']

        return self._execute(request, response_handler)

    def validate(self, query):
        """Parse and validate the query without executing it.

        :param query: Query to validate.
        :type query: str | unicode
        :return: Query details.
        :rtype: dict
        :raise arango.exceptions.AQLQueryValidateError: If validation fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/query',
            data={'query': query}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryValidateError(resp, request)
            body = resp.body
            body.pop('code', None)
            body.pop('error', None)
            if 'bindVars' in body:
                body['bind_vars'] = body.pop('bindVars')
            return body

        return self._execute(request, response_handler)

    def execute(self,
                query,
                count=False,
                batch_size=None,
                ttl=None,
                bind_vars=None,
                full_count=None,
                max_plans=None,
                optimizer_rules=None,
                cache=None,
                memory_limit=0,
                fail_on_warning=None,
                profile=None,
                max_transaction_size=None,
                max_warning_count=None,
                intermediate_commit_count=None,
                intermediate_commit_size=None,
                satellite_sync_wait=None,
                read_collections=None,
                write_collections=None,
                stream=None,
                skip_inaccessible_cols=None,
                max_runtime=None):
        """Execute the query and return the result cursor.

        :param query: Query to execute.
        :type query: str | unicode
        :param count: If set to True, the total document count is included in
            the result cursor.
        :type count: bool
        :param batch_size: Number of documents fetched by the cursor in one
            round trip.
        :type batch_size: int
        :param ttl: Server side time-to-live for the cursor in seconds.
        :type ttl: int
        :param bind_vars: Bind variables for the query.
        :type bind_vars: dict
        :param full_count: This parameter applies only to queries with LIMIT
            clauses. If set to True, the number of matched documents before
            the last LIMIT clause executed is included in teh cursor. This is
            similar to MySQL SQL_CALC_FOUND_ROWS hint. Using this disables a
            few LIMIT optimizations and may lead to a longer query execution.
        :type full_count: bool
        :param max_plans: Max number of plans the optimizer generates.
        :type max_plans: int
        :param optimizer_rules: List of optimizer rules.
        :type optimizer_rules: [str | unicode]
        :param cache: If set to True, the query cache is used. The operation
            mode of the query cache must be set to "on" or "demand".
        :type cache: bool
        :param memory_limit: Max amount of memory the query is allowed to use
            in bytes. If the query goes over the limit, it fails with error
            "resource limit exceeded". Value 0 indicates no limit.
        :type memory_limit: int
        :param fail_on_warning: If set to True, the query throws an exception
            instead of producing a warning. This parameter can be used during
            development to catch issues early. If set to False, warnings are
            returned with the query result. There is a server configuration
            option "--query.fail-on-warning" for setting the default value for
            this behaviour so it does not need to be set per-query.
        :type fail_on_warning: bool
        :param profile: Return additional profiling details in the cursor,
            unless the query cache is used.
        :type profile: bool
        :param max_transaction_size: Transaction size limit in bytes. Applies
            only to RocksDB storage engine.
        :type max_transaction_size: int
        :param max_warning_count: Max number of warnings returned.
        :type max_warning_count: int
        :param intermediate_commit_count: Max number of operations after
            which an intermediate commit is performed automatically. Applies
            only to RocksDB storage engine.
        :type intermediate_commit_count: int
        :param intermediate_commit_size: Max size of operations in bytes after
            which an intermediate commit is performed automatically. Applies
            only to RocksDB storage engine.
        :type intermediate_commit_size: int
        :param satellite_sync_wait: Number of seconds in which the server must
            synchronize the satellite collections involved in the query. When
            the threshold is reached, the query is stopped. Available only for
            enterprise version of ArangoDB.
        :type satellite_sync_wait: int | float
        :param read_collections: Names of collections read during query
            execution. This parameter is deprecated.
        :type read_collections: [str | unicode]
        :param write_collections: Names of collections written to during query
            execution. This parameter is deprecated.
        :type write_collections: [str | unicode]
        :param stream: If set to True, query is executed in streaming fashion:
            query result is not stored server-side but calculated on the fly.
            Note: long-running queries hold collection locks for as long as the
            cursor exists. If set to False, query is executed right away in its
            entirety. Results are either returned right away (if the result set
            is small enough), or stored server-side and accessible via cursors
            (while respecting the ttl). You should use this parameter only for
            short-running queries or without exclusive locks (write-locks on
            MMFiles). Note: parameters **cache**, **count** and **full_count**
            do not work for streaming queries. Query statistics, warnings and
            profiling data are made available only after the query is finished.
            Default value is False.
        :type stream: bool
        :param skip_inaccessible_cols: If set to True, collections without user
            access are skipped, and query executes normally instead of raising
            an error. This helps certain use cases: a graph may contain several
            collections, and users with different access levels may execute the
            same query. This parameter lets you limit the result set by user
            access. Cannot be used in :doc:`transactions <transaction>` and is
            available only for enterprise version of ArangoDB. Default value is
            False.
        :type skip_inaccessible_cols: bool
        :param max_runtime: Query must be executed within this given timeout or
            it is killed. The value is specified in seconds. Default value
            is 0.0 (no timeout).
        :type max_runtime: int | float
        :return: Result cursor.
        :rtype: arango.cursor.Cursor
        :raise arango.exceptions.AQLQueryExecuteError: If execute fails.
        """
        data = {'query': query, 'count': count}
        if batch_size is not None:
            data['batchSize'] = batch_size
        if ttl is not None:
            data['ttl'] = ttl
        if bind_vars is not None:
            data['bindVars'] = bind_vars
        if cache is not None:
            data['cache'] = cache
        if memory_limit is not None:
            data['memoryLimit'] = memory_limit

        options = {}
        if full_count is not None:
            options['fullCount'] = full_count
        if max_plans is not None:
            options['maxNumberOfPlans'] = max_plans
        if optimizer_rules is not None:
            options['optimizer'] = {'rules': optimizer_rules}
        if fail_on_warning is not None:
            options['failOnWarning'] = fail_on_warning
        if profile is not None:
            options['profile'] = profile
        if max_transaction_size is not None:
            options['maxTransactionSize'] = max_transaction_size
        if max_warning_count is not None:
            options['maxWarningCount'] = max_warning_count
        if intermediate_commit_count is not None:
            options['intermediateCommitCount'] = intermediate_commit_count
        if intermediate_commit_size is not None:
            options['intermediateCommitSize'] = intermediate_commit_size
        if satellite_sync_wait is not None:
            options['satelliteSyncWait'] = satellite_sync_wait
        if stream is not None:
            options['stream'] = stream
        if skip_inaccessible_cols is not None:
            options['skipInaccessibleCollections'] = skip_inaccessible_cols
        if max_runtime is not None:
            options['maxRuntime'] = max_runtime

        if options:
            data['options'] = options
        data.update(options)

        request = Request(
            method='post',
            endpoint='/_api/cursor',
            data=data,
            read=read_collections,
            write=write_collections
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryExecuteError(resp, request)
            return Cursor(self._conn, resp.body)

        return self._execute(request, response_handler)

    async def kill(self, query_id):
        """Kill a running query.

        :param query_id: Query ID.
        :type query_id: str | unicode
        :return: True if kill request was sent successfully.
        :rtype: bool
        :raise arango.exceptions.AQLQueryKillError: If the send fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/query/{}'.format(query_id)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryKillError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def queries(self):
        """Return the currently running AQL queries.

        :return: Running AQL queries.
        :rtype: [dict]
        :raise arango.exceptions.AQLQueryListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/query/current'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryListError(resp, request)
            return [format_aql_query(q) for q in resp.body]

        return await self._execute(request, response_handler)

    async def slow_queries(self):
        """Return a list of all slow AQL queries.

        :return: Slow AQL queries.
        :rtype: [dict]
        :raise arango.exceptions.AQLQueryListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/query/slow'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryListError(resp, request)
            return [format_aql_query(q) for q in resp.body]

        return await self._execute(request, response_handler)

    async def clear_slow_queries(self):
        """Clear slow AQL queries.

        :return: True if slow queries were cleared successfully.
        :rtype: bool
        :raise arango.exceptions.AQLQueryClearError: If operation fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/query/slow'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryClearError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def tracking(self):
        """Return AQL query tracking properties.

        :return: AQL query tracking properties.
        :rtype: dict
        :raise arango.exceptions.AQLQueryTrackingGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/query/properties'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryTrackingGetError(resp, request)
            return format_aql_tracking(resp.body)

        return await self._execute(request, response_handler)

    async def set_tracking(
            self,
            enabled=None,
            max_slow_queries=None,
            slow_query_threshold=None,
            max_query_string_length=None,
            track_bind_vars=None,
            track_slow_queries=None):
        """Configure AQL query tracking properties

        :return: Updated AQL query tracking properties.
        :rtype: dict
        :raise arango.exceptions.AQLQueryTrackingSetError: If operation fails.
        """
        data = {}
        if enabled is not None:
            data['enabled'] = enabled
        if max_slow_queries is not None:
            data['maxSlowQueries'] = max_slow_queries
        if max_query_string_length is not None:
            data['maxQueryStringLength'] = max_query_string_length
        if slow_query_threshold is not None:
            data['slowQueryThreshold'] = slow_query_threshold
        if track_bind_vars is not None:
            data['trackBindVars'] = track_bind_vars
        if track_slow_queries is not None:
            data['trackSlowQueries'] = track_slow_queries

        request = Request(
            method='put',
            endpoint='/_api/query/properties',
            data=data
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLQueryTrackingSetError(resp, request)
            return format_aql_tracking(resp.body)

        return await self._execute(request, response_handler)

    async def functions(self):
        """List the AQL functions defined in the database.

        :return: AQL functions.
        :rtype: [dict]
        :raise arango.exceptions.AQLFunctionListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/aqlfunction'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLFunctionListError(resp, request)

            body = resp.body or {}

            if 'result' not in body:  # pragma: no cover
                return []

            for item in body['result']:
                if 'isDeterministic' in item:
                    item['is_deterministic'] = item.pop('isDeterministic')

            return body['result']

        return await self._execute(request, response_handler)

    async def create_function(self, name, code):
        """Create a new AQL function.

        :param name: AQL function name.
        :type name: str | unicode
        :param code: Function definition in Javascript.
        :type code: str | unicode
        :return: Whether the AQL function was newly created or an existing one
            was replaced.
        :rtype: dict
        :raise arango.exceptions.AQLFunctionCreateError: If create fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/aqlfunction',
            data={'name': name, 'code': code}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLFunctionCreateError(resp, request)
            return {'is_new': resp.body['isNewlyCreated']}

        return await self._execute(request, response_handler)

    async def delete_function(self, name, group=False, ignore_missing=False):
        """Delete an AQL function.

        :param name: AQL function name.
        :type name: str | unicode
        :param group: If set to True, value of parameter **name** is treated
            as a namespace prefix, and all functions in the namespace are
            deleted. If set to False, the value of **name** must be a fully
            qualified function name including any namespaces.
        :type group: bool
        :param ignore_missing: Do not raise an exception on missing function.
        :type ignore_missing: bool
        :return: Number of AQL functions deleted if operation was successful,
            False if function(s) was not found and **ignore_missing** was set
            to True.
        :rtype: dict | bool
        :raise arango.exceptions.AQLFunctionDeleteError: If delete fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/aqlfunction/{}'.format(name),
            params={'group': group}
        )

        def response_handler(resp):
            if resp.error_code == 1582 and ignore_missing:
                return False
            if not resp.is_success:
                raise AQLFunctionDeleteError(resp, request)
            return {'deleted': resp.body['deletedCount']}

        return await self._execute(request, response_handler)


class AQLQueryCache(APIWrapper):
    """AQL Query Cache API wrapper."""

    def __repr__(self):
        return '<AQLQueryCache in {}>'.format(self._conn.db_name)

    async def properties(self):
        """Return the query cache properties.

        :return: Query cache properties.
        :rtype: dict
        :raise arango.exceptions.AQLCachePropertiesError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/query-cache/properties'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLCachePropertiesError(resp, request)
            return format_aql_cache(resp.body)

        return await self._execute(request, response_handler)

    async def configure(
            self,
            mode=None,
            max_results=None,
            max_results_size=None,
            max_entry_size=None,
            include_system=None):
        """Configure the query cache properties.

        :param mode: Operation mode. Allowed values are "off", "on" and
            "demand".
        :type mode: str | unicode
        :param max_results: Max number of query results stored per
            database-specific cache.
        :type max_results: int
        :param max_results_size: Max cumulative size of query results stored
            per database-specific cache.
        :type max_results_size: int
        :param max_entry_size: Max entry size of each query result stored per
            database-specific cache.
        :type max_entry_size: int
        :param include_system: Store results of queries in system collections.
        :type include_system: bool
        :return: Query cache properties.
        :rtype: dict
        :raise arango.exceptions.AQLCacheConfigureError: If operation fails.
        """
        data = {}
        if mode is not None:
            data['mode'] = mode
        if max_results is not None:
            data['maxResults'] = max_results
        if max_results_size is not None:
            data['maxResultsSize'] = max_results_size
        if max_entry_size is not None:
            data['maxEntrySize'] = max_entry_size
        if include_system is not None:
            data['includeSystem'] = include_system

        request = Request(
            method='put',
            endpoint='/_api/query-cache/properties',
            data=data
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLCacheConfigureError(resp, request)
            return format_aql_cache(resp.body)

        return await self._execute(request, response_handler)

    async def entries(self):
        """Return the query cache entries.

        :return: Query cache entries.
        :rtype: list
        :raise AQLCacheEntriesError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/query-cache/entries'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLCacheEntriesError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def clear(self):
        """Clear the query cache.

        :return: True if query cache was cleared successfully.
        :rtype: bool
        :raise arango.exceptions.AQLCacheClearError: If operation fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/query-cache'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise AQLCacheClearError(resp, request)
            return True

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

from six import string_types

import pytest
from aioarangodb.collection import EdgeCollection
from aioarangodb.exceptions import (
    DocumentDeleteError,
    DocumentGetError,
    DocumentInsertError,
    DocumentParseError,
    DocumentReplaceError,
    DocumentRevisionError,
    DocumentUpdateError,
    EdgeDefinitionListError,
    EdgeDefinitionCreateError,
    EdgeDefinitionDeleteError,
    EdgeDefinitionReplaceError,
    EdgeListError,
    GraphListError,
    GraphCreateError,
    GraphDeleteError,
    GraphPropertiesError,
    GraphTraverseError,
    VertexCollectionCreateError,
    VertexCollectionDeleteError,
    VertexCollectionListError
)
from aioarangodb.tests.helpers import (
    assert_raises,
    clean_doc,
    empty_collection,
    extract,
    generate_col_name,
    generate_graph_name,
    generate_doc_key,
)
pytestmark = pytest.mark.asyncio


async def test_graph_properties(graph, bad_graph, db):
    assert repr(graph) == '<Graph {}>'.format(graph.name)

    properties = await graph.properties()

    assert properties['id'] == '_graphs/{}'.format(graph.name)
    assert properties['name'] == graph.name
    assert len(properties['edge_definitions']) == 1
    assert 'orphan_collections' in properties
    assert isinstance(properties['revision'], string_types)

    # Test properties with bad database
    with assert_raises(GraphPropertiesError):
        await bad_graph.properties()

    new_graph_name = generate_graph_name()
    new_graph = await db.create_graph(new_graph_name)
    properties = await new_graph.properties()
    assert properties['id'] == '_graphs/{}'.format(new_graph_name)
    assert properties['name'] == new_graph_name
    assert properties['edge_definitions'] == []
    assert properties['orphan_collections'] == []
    assert isinstance(properties['revision'], string_types)


async def test_graph_management(db, bad_db):
    # Test create graph
    graph_name = generate_graph_name()
    assert await db.has_graph(graph_name) is False

    graph = await db.create_graph(graph_name)
    assert await db.has_graph(graph_name) is True
    assert graph.name == graph_name
    assert graph.db_name == db.name

    # Test create duplicate graph
    with assert_raises(GraphCreateError) as err:
        await db.create_graph(graph_name)
    assert err.value.error_code == 1925

    # Test get graph
    result = db.graph(graph_name)
    assert result.name == graph.name
    assert result.db_name == graph.db_name

    # Test get graphs
    result = await db.graphs()
    for entry in result:
        assert 'revision' in entry
        assert 'edge_definitions' in entry
        assert 'orphan_collections' in entry
    assert graph_name in await extract('name', await db.graphs())

    # Test get graphs with bad database
    with assert_raises(GraphListError) as err:
        await bad_db.graphs()
    assert err.value.error_code in {11, 1228}

    # Test delete graph
    assert await db.delete_graph(graph_name) is True
    assert graph_name not in await extract('name', await db.graphs())

    # Test delete missing graph
    with assert_raises(GraphDeleteError) as err:
        await db.delete_graph(graph_name)
    assert err.value.error_code == 1924
    assert await db.delete_graph(graph_name, ignore_missing=True) is False

    # Create a graph with vertex and edge collections and delete the graph
    graph = await db.create_graph(graph_name)
    ecol_name = generate_col_name()
    fvcol_name = generate_col_name()
    tvcol_name = generate_col_name()

    await graph.create_vertex_collection(fvcol_name)
    await graph.create_vertex_collection(tvcol_name)
    await graph.create_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[fvcol_name],
        to_vertex_collections=[tvcol_name]
    )
    collections = await extract('name', await db.collections())
    assert fvcol_name in collections
    assert tvcol_name in collections
    assert ecol_name in collections

    await db.delete_graph(graph_name)
    collections = await extract('name', await db.collections())
    assert fvcol_name in collections
    assert tvcol_name in collections
    assert ecol_name in collections

    # Create a graph with vertex and edge collections and delete all
    graph = await db.create_graph(graph_name)
    await graph.create_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[fvcol_name],
        to_vertex_collections=[tvcol_name]
    )
    await db.delete_graph(graph_name, drop_collections=True)
    collections = await extract('name', await db.collections())
    assert fvcol_name not in collections
    assert tvcol_name not in collections
    assert ecol_name not in collections


async def test_vertex_collection_management(db, graph, bad_graph):
    # Test create valid "from" vertex collection
    fvcol_name = generate_col_name()
    assert not await graph.has_vertex_collection(fvcol_name)
    assert not await db.has_collection(fvcol_name)

    fvcol = await graph.create_vertex_collection(fvcol_name)
    assert await graph.has_vertex_collection(fvcol_name)
    assert await db.has_collection(fvcol_name)
    assert fvcol.name == fvcol_name
    assert fvcol.graph == graph.name
    assert fvcol_name in repr(fvcol)
    assert fvcol_name in await graph.vertex_collections()
    assert fvcol_name in await extract('name', await db.collections())

    # Test create duplicate vertex collection
    with assert_raises(VertexCollectionCreateError) as err:
        await graph.create_vertex_collection(fvcol_name)
    assert err.value.error_code == 1938
    assert fvcol_name in await graph.vertex_collections()
    assert fvcol_name in await extract('name', await db.collections())

    # Test create valid "to" vertex collection
    tvcol_name = generate_col_name()
    assert not await graph.has_vertex_collection(tvcol_name)
    assert not await db.has_collection(tvcol_name)

    tvcol = await graph.create_vertex_collection(tvcol_name)
    assert await graph.has_vertex_collection(tvcol_name)
    assert await db.has_collection(tvcol_name)
    assert tvcol_name == tvcol_name
    assert tvcol.graph == graph.name
    assert tvcol_name in repr(tvcol)
    assert tvcol_name in await graph.vertex_collections()
    assert tvcol_name in await extract('name', await db.collections())

    # Test list vertex collection via bad database
    with assert_raises(VertexCollectionListError) as err:
        await bad_graph.vertex_collections()
    assert err.value.error_code in {11, 1228}

    # Test delete missing vertex collection
    with assert_raises(VertexCollectionDeleteError) as err:
        await graph.delete_vertex_collection(generate_col_name())
    assert err.value.error_code in {1926, 1928}

    # Test delete "to" vertex collection with purge option
    assert await graph.delete_vertex_collection(tvcol_name, purge=True) is True
    assert tvcol_name not in await graph.vertex_collections()
    assert fvcol_name in await extract('name', await db.collections())
    assert tvcol_name not in await extract('name', await db.collections())
    assert not await graph.has_vertex_collection(tvcol_name)

    # Test delete "from" vertex collection without purge option
    assert await graph.delete_vertex_collection(fvcol_name, purge=False) is True
    assert fvcol_name not in await graph.vertex_collections()
    assert fvcol_name in await extract('name', await db.collections())
    assert not await graph.has_vertex_collection(fvcol_name)


async def test_edge_definition_management(db, graph, bad_graph):
    ecol_name = generate_col_name()
    assert not await graph.has_edge_definition(ecol_name)
    assert not await graph.has_edge_collection(ecol_name)
    assert not await db.has_collection(ecol_name)

    # Test create edge definition with existing vertex collections
    fvcol_name = generate_col_name()
    tvcol_name = generate_col_name()
    ecol_name = generate_col_name()
    ecol = await graph.create_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[fvcol_name],
        to_vertex_collections=[tvcol_name]
    )
    assert ecol.name == ecol_name
    assert ecol.graph == graph.name
    assert repr(ecol) == '<EdgeCollection {}>'.format(ecol.name)
    assert {
        'edge_collection': ecol_name,
        'from_vertex_collections': [fvcol_name],
        'to_vertex_collections': [tvcol_name]
    } in await graph.edge_definitions()
    assert ecol_name in await extract('name', await db.collections())

    vertex_collections = await graph.vertex_collections()
    assert fvcol_name in vertex_collections
    assert tvcol_name in vertex_collections

    # Test create duplicate edge definition
    with assert_raises(EdgeDefinitionCreateError) as err:
        await graph.create_edge_definition(
            edge_collection=ecol_name,
            from_vertex_collections=[fvcol_name],
            to_vertex_collections=[tvcol_name]
        )
    assert err.value.error_code == 1920

    # Test create edge definition with missing vertex collection
    bad_vcol_name = generate_col_name()
    ecol_name = generate_col_name()
    ecol = await graph.create_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[bad_vcol_name],
        to_vertex_collections=[bad_vcol_name]
    )
    assert await graph.has_edge_definition(ecol_name)
    assert await graph.has_edge_collection(ecol_name)
    assert ecol.name == ecol_name
    assert {
        'edge_collection': ecol_name,
        'from_vertex_collections': [bad_vcol_name],
        'to_vertex_collections': [bad_vcol_name]
    } in await graph.edge_definitions()
    assert bad_vcol_name in await graph.vertex_collections()
    assert bad_vcol_name in await extract('name', await db.collections())
    assert bad_vcol_name in await extract('name', await db.collections())

    # Test list edge definition with bad database
    with assert_raises(EdgeDefinitionListError) as err:
        await bad_graph.edge_definitions()
    assert err.value.error_code in {11, 1228}

    # Test replace edge definition (happy path)
    ecol = await graph.replace_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[tvcol_name],
        to_vertex_collections=[fvcol_name]
    )
    assert isinstance(ecol, EdgeCollection)
    assert ecol.name == ecol_name
    assert {
        'edge_collection': ecol_name,
        'from_vertex_collections': [tvcol_name],
        'to_vertex_collections': [fvcol_name]
    } in await graph.edge_definitions()

    # Test replace missing edge definition
    bad_ecol_name = generate_col_name()
    with assert_raises(EdgeDefinitionReplaceError):
        await graph.replace_edge_definition(
            edge_collection=bad_ecol_name,
            from_vertex_collections=[],
            to_vertex_collections=[fvcol_name]
        )

    # Test delete missing edge definition
    with assert_raises(EdgeDefinitionDeleteError) as err:
        await graph.delete_edge_definition(bad_ecol_name)
    assert err.value.error_code == 1930

    # Test delete existing edge definition with purge
    assert await graph.delete_edge_definition(ecol_name, purge=True) is True
    assert ecol_name not in \
        await extract('edge_collection', await graph.edge_definitions())
    assert not await graph.has_edge_definition(ecol_name)
    assert not await graph.has_edge_collection(ecol_name)
    assert ecol_name not in await extract('name', await db.collections())


async def test_create_graph_with_edge_definition(db):
    new_graph_name = generate_graph_name()
    new_ecol_name = generate_col_name()
    fvcol_name = generate_col_name()
    tvcol_name = generate_col_name()
    ovcol_name = generate_col_name()

    edge_definition = {
        'edge_collection': new_ecol_name,
        'from_vertex_collections': [fvcol_name],
        'to_vertex_collections': [tvcol_name]
    }
    new_graph = await db.create_graph(
        new_graph_name,
        edge_definitions=[edge_definition],
        orphan_collections=[ovcol_name]
    )
    assert edge_definition in await new_graph.edge_definitions()


async def test_vertex_management(fvcol, bad_fvcol, fvdocs):
    # Test insert vertex with no key
    result = await fvcol.insert({})
    assert await fvcol.has(result['_key'])
    assert await fvcol.count() == 1
    await empty_collection(fvcol)

    # Test insert vertex with ID
    vertex_id = fvcol.name + '/' + 'foo'
    await fvcol.insert({'_id': vertex_id})
    assert await fvcol.has('foo')
    assert await fvcol.has(vertex_id)
    assert await fvcol.count() == 1
    await empty_collection(fvcol)

    # Test insert vertex with return_new set to True
    result = await fvcol.insert({'_id': vertex_id}, return_new=True)
    assert 'new' in result
    assert 'vertex' in result
    assert await fvcol.count() == 1
    await empty_collection(fvcol)

    with assert_raises(DocumentParseError) as err:
        await fvcol.insert({'_id': generate_col_name() + '/' + 'foo'})
    assert 'bad collection name' in err.value.message

    vertex = fvdocs[0]
    key = vertex['_key']

    # Test insert first valid vertex
    result = await fvcol.insert(vertex, sync=True)
    assert result['_key'] == key
    assert '_rev' in result
    assert await fvcol.has(vertex) and await fvcol.has(key)
    assert await fvcol.count() == 1
    assert (await fvcol.get(key))['val'] == vertex['val']

    # Test insert duplicate vertex
    with assert_raises(DocumentInsertError) as err:
        await fvcol.insert(vertex)
    assert err.value.error_code in {1202, 1210}
    assert await fvcol.count() == 1

    vertex = fvdocs[1]
    key = vertex['_key']

    # Test insert second valid vertex
    result = await fvcol.insert(vertex)
    assert result['_key'] == key
    assert '_rev' in result
    assert await fvcol.has(vertex) and await fvcol.has(key)
    assert await fvcol.count() == 2
    assert (await fvcol.get(key))['val'] == vertex['val']

    vertex = fvdocs[2]
    key = vertex['_key']

    # Test insert third valid vertex with silent set to True
    assert await fvcol.insert(vertex, silent=True) is True
    assert await fvcol.count() == 3
    assert (await fvcol.get(key))['val'] == vertex['val']

    # Test get missing vertex
    if fvcol.context != 'transaction':
        assert await fvcol.get(generate_doc_key()) is None

    # Test get existing edge by body with "_key" field
    result = await fvcol.get({'_key': key})
    assert await clean_doc(result) == vertex

    # Test get existing edge by body with "_id" field
    result = await fvcol.get({'_id': fvcol.name + '/' + key})
    assert await clean_doc(result) == vertex

    # Test get existing vertex by key
    result = await fvcol.get(key)
    assert await clean_doc(result) == vertex

    # Test get existing vertex by ID
    result = await fvcol.get(fvcol.name + '/' + key)
    assert await clean_doc(result) == vertex

    # Test get existing vertex with bad revision
    old_rev = result['_rev']
    with assert_raises(DocumentRevisionError) as err:
        await fvcol.get(key, rev=old_rev + '1', check_rev=True)
    assert err.value.error_code in {1903, 1200}

    # Test get existing vertex with bad database
    with assert_raises(DocumentGetError) as err:
        await bad_fvcol.get(key)
    assert err.value.error_code in {11, 1228}

    # Test update vertex with a single field change
    assert 'foo' not in await fvcol.get(key)
    result = await fvcol.update({'_key': key, 'foo': 100})
    assert result['_key'] == key
    assert (await fvcol.get(key))['foo'] == 100
    old_rev = (await fvcol.get(key))['_rev']

    # Test update vertex with return_new and return_old set to True
    result = await fvcol.update(
        {'_key': key, 'foo': 100},
        return_old=True,
        return_new=True
    )
    assert 'old' in result
    assert 'new' in result
    assert 'vertex' in result

    # Test update vertex with silent set to True
    assert 'bar' not in await fvcol.get(vertex)
    assert await fvcol.update({'_key': key, 'bar': 200}, silent=True) is True
    assert (await fvcol.get(vertex))['bar'] == 200
    assert (await fvcol.get(vertex))['_rev'] != old_rev
    old_rev = (await fvcol.get(key))['_rev']

    # Test update vertex with multiple field changes
    result = await fvcol.update({'_key': key, 'foo': 200, 'bar': 300})
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await fvcol.get(key))['foo'] == 200
    assert (await fvcol.get(key))['bar'] == 300
    old_rev = result['_rev']

    # Test update vertex with correct revision
    result = await fvcol.update({'_key': key, '_rev': old_rev, 'bar': 400})
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await fvcol.get(key))['foo'] == 200
    assert (await fvcol.get(key))['bar'] == 400
    old_rev = result['_rev']

    # Test update vertex with bad revision
    if fvcol.context != 'transaction':
        new_rev = old_rev + '1'
        with assert_raises(DocumentRevisionError) as err:
            await fvcol.update({'_key': key, '_rev': new_rev, 'bar': 500})
        assert err.value.error_code in {1200, 1903}
        assert (await fvcol.get(key))['foo'] == 200
        assert (await fvcol.get(key))['bar'] == 400

    # Test update vertex in missing vertex collection
    with assert_raises(DocumentUpdateError) as err:
        await bad_fvcol.update({'_key': key, 'bar': 500})
    assert err.value.error_code in {11, 1228}
    assert (await fvcol.get(key))['foo'] == 200
    assert (await fvcol.get(key))['bar'] == 400

    # Test update vertex with sync set to True
    result = await fvcol.update({'_key': key, 'bar': 500}, sync=True)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await fvcol.get(key))['foo'] == 200
    assert (await fvcol.get(key))['bar'] == 500
    old_rev = result['_rev']

    # Test update vertex with keep_none set to True
    result = await fvcol.update({'_key': key, 'bar': None}, keep_none=True)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await fvcol.get(key))['foo'] == 200
    assert (await fvcol.get(key))['bar'] is None
    old_rev = result['_rev']

    # Test update vertex with keep_none set to False
    result = await fvcol.update({'_key': key, 'foo': None}, keep_none=False)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert 'foo' not in (await fvcol.get(key))
    assert (await fvcol.get(key))['bar'] is None

    # Test replace vertex with a single field change
    result = await fvcol.replace({'_key': key, 'baz': 100})
    assert result['_key'] == key
    assert 'foo' not in await fvcol.get(key)
    assert 'bar' not in await fvcol.get(key)
    assert (await fvcol.get(key))['baz'] == 100
    old_rev = result['_rev']

    # Test replace vertex with return_new and return_old set to True
    result = await fvcol.replace(
        {'_key': key, 'baz': 100},
        return_old=True,
        return_new=True
    )
    assert 'old' in result
    assert 'new' in result
    assert 'vertex' in result

    # Test replace vertex with silent set to True
    assert await fvcol.replace({'_key': key, 'bar': 200}, silent=True) is True
    assert 'foo' not in await fvcol.get(key)
    assert 'baz' not in await fvcol.get(vertex)
    assert (await fvcol.get(vertex))['bar'] == 200
    assert await fvcol.count() == 3
    assert (await fvcol.get(vertex))['_rev'] != old_rev
    old_rev = (await fvcol.get(vertex))['_rev']

    # Test replace vertex with multiple field changes
    vertex = {'_key': key, 'foo': 200, 'bar': 300}
    result = await fvcol.replace(vertex)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert await clean_doc(await fvcol.get(key)) == vertex
    old_rev = result['_rev']

    # Test replace vertex with correct revision
    vertex = {'_key': key, '_rev': old_rev, 'bar': 500}
    result = await fvcol.replace(vertex)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert await clean_doc(await fvcol.get(key)) == await clean_doc(vertex)
    old_rev = result['_rev']

    # Test replace vertex with bad revision
    if fvcol.context != 'transaction':
        new_rev = old_rev + '10'
        vertex = {'_key': key, '_rev': new_rev, 'bar': 600}
        with assert_raises(DocumentRevisionError, DocumentReplaceError) as err:
            await fvcol.replace(vertex)
        assert err.value.error_code in {1200, 1903}
        assert (await fvcol.get(key))['bar'] == 500
        assert 'foo' not in await fvcol.get(key)

    # Test replace vertex with bad database
    with assert_raises(DocumentReplaceError) as err:
        await bad_fvcol.replace({'_key': key, 'bar': 600})
    assert err.value.error_code in {11, 1228}
    assert (await fvcol.get(key))['bar'] == 500
    assert 'foo' not in await fvcol.get(key)

    # Test replace vertex with sync set to True
    vertex = {'_key': key, 'bar': 400, 'foo': 200}
    result = await fvcol.replace(vertex, sync=True)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await fvcol.get(key))['foo'] == 200
    assert (await fvcol.get(key))['bar'] == 400

    # Test delete vertex with bad revision
    if fvcol.context != 'transaction':
        old_rev = (await fvcol.get(key))['_rev']
        vertex['_rev'] = old_rev + '1'
        with assert_raises(DocumentRevisionError, DocumentDeleteError) as err:
            await fvcol.delete(vertex, check_rev=True)
        assert err.value.error_code in {1200, 1903}
        vertex['_rev'] = old_rev
        assert await fvcol.has(vertex)

    # Test delete missing vertex
    bad_key = generate_doc_key()
    with assert_raises(DocumentDeleteError) as err:
        await fvcol.delete(bad_key, ignore_missing=False)
    assert err.value.error_code == 1202
    if fvcol.context != 'transaction':
        assert await fvcol.delete(bad_key, ignore_missing=True) is False

    # Test delete existing vertex with sync set to True
    assert await fvcol.delete(vertex, sync=True, check_rev=False) is True
    if fvcol.context != 'transaction':
        assert await fvcol.get(vertex) is None
    assert not await fvcol.has(vertex)
    assert await fvcol.count() == 2

    # Test delete existing vertex with return_old set to True
    vertex = fvdocs[1]
    result = await fvcol.delete(vertex, return_old=True)
    assert 'old' in result
    assert await fvcol.count() == 1
    await empty_collection(fvcol)


async def test_vertex_management_via_graph(graph, fvcol):
    # Test insert vertex via graph object
    result = await graph.insert_vertex(fvcol.name, {})
    assert await fvcol.has(result['_key'])
    assert await fvcol.count() == 1
    vertex_id = result['_id']

    # Test get vertex via graph object
    assert (await graph.vertex(vertex_id))['_id'] == vertex_id

    # Test update vertex via graph object
    result = await graph.update_vertex({'_id': vertex_id, 'foo': 100})
    assert result['_id'] == vertex_id
    assert (await fvcol.get(vertex_id))['foo'] == 100

    # Test replace vertex via graph object
    result = await graph.replace_vertex({'_id': vertex_id, 'bar': 200})
    assert result['_id'] == vertex_id
    assert 'foo' not in await fvcol.get(vertex_id)
    assert (await fvcol.get(vertex_id))['bar'] == 200

    # Test delete vertex via graph object
    assert await graph.delete_vertex(vertex_id) is True
    assert not await fvcol.get(vertex_id)
    assert await fvcol.count() == 0


async def test_edge_management(ecol, bad_ecol, edocs, fvcol, fvdocs, tvcol, tvdocs):
    for vertex in fvdocs:
        await fvcol.insert(vertex)
    for vertex in tvdocs:
        await tvcol.insert(vertex)

    edge = edocs[0]
    key = edge['_key']

    # Test insert edge with no key
    no_key_edge = {'_from': edge['_from'], '_to': edge['_to']}
    result = await ecol.insert(no_key_edge)
    assert await ecol.has(result['_key'])
    assert await ecol.count() == 1
    await empty_collection(ecol)

    # Test insert edge with return_new set to True
    result = await ecol.insert(no_key_edge, return_new=True)
    assert 'new' in result
    assert await ecol.has(result['edge']['_key'])
    assert await ecol.count() == 1
    await empty_collection(ecol)

    # Test insert vertex with ID
    edge_id = ecol.name + '/' + 'foo'
    await ecol.insert({
        '_id': edge_id,
        '_from': edge['_from'],
        '_to': edge['_to']
    })
    assert await ecol.has('foo')
    assert await ecol.has(edge_id)
    assert await ecol.count() == 1
    await empty_collection(ecol)

    with assert_raises(DocumentParseError) as err:
        await ecol.insert({
            '_id': generate_col_name() + '/' + 'foo',
            '_from': edge['_from'],
            '_to': edge['_to']
        })
    assert 'bad collection name' in err.value.message

    # Test insert first valid edge
    result = await ecol.insert(edge)
    assert result['_key'] == key
    assert '_rev' in result
    assert await ecol.has(edge) and await ecol.has(key)
    assert await ecol.count() == 1
    assert (await ecol.get(key))['_from'] == edge['_from']
    assert (await ecol.get(key))['_to'] == edge['_to']

    # Test insert duplicate edge
    with assert_raises(DocumentInsertError) as err:
        assert await ecol.insert(edge)
    assert err.value.error_code in {1202, 1210, 1906}
    assert await ecol.count() == 1

    edge = edocs[1]
    key = edge['_key']

    # Test insert second valid edge with silent set to True
    assert await ecol.insert(edge, sync=True, silent=True) is True
    assert await ecol.has(edge) and await ecol.has(key)
    assert await ecol.count() == 2
    assert (await ecol.get(key))['_from'] == edge['_from']
    assert (await ecol.get(key))['_to'] == edge['_to']

    # Test insert third valid edge using link method
    from_vertex = await fvcol.get(fvdocs[2])
    to_vertex = await tvcol.get(tvdocs[2])
    result = await ecol.link(from_vertex, to_vertex, sync=False)
    assert await ecol.has(result['_key'])
    assert await ecol.count() == 3

    # Test insert fourth valid edge using link method
    from_vertex = await fvcol.get(fvdocs[2])
    to_vertex = await tvcol.get(tvdocs[0])
    assert await ecol.link(
        from_vertex['_id'],
        to_vertex['_id'],
        {'_id': ecol.name + '/foo'},
        sync=True,
        silent=True
    ) is True
    assert await ecol.has('foo')
    assert await ecol.count() == 4

    with assert_raises(DocumentParseError) as err:
        assert await ecol.link({}, {})
    assert err.value.message == 'field "_id" required'

    # Test get missing vertex
    bad_document_key = generate_doc_key()
    if ecol.context != 'transaction':
        assert await ecol.get(bad_document_key) is None

    # Test get existing edge by body with "_key" field
    result = await ecol.get({'_key': key})
    assert await clean_doc(result) == edge

    # Test get existing edge by body with "_id" field
    result = await ecol.get({'_id': ecol.name + '/' + key})
    assert await clean_doc(result) == edge

    # Test get existing edge by key
    result = await ecol.get(key)
    assert await clean_doc(result) == edge

    # Test get existing edge by ID
    result = await ecol.get(ecol.name + '/' + key)
    assert await clean_doc(result) == edge

    # Test get existing edge with bad revision
    old_rev = result['_rev']
    with assert_raises(DocumentRevisionError) as err:
        await ecol.get(key, rev=old_rev + '1')
    assert err.value.error_code in {1903, 1200}

    # Test get existing edge with bad database
    with assert_raises(DocumentGetError) as err:
        await bad_ecol.get(key)
    assert err.value.error_code in {11, 1228}

    # Test update edge with a single field change
    assert 'foo' not in await ecol.get(key)
    result = await ecol.update({'_key': key, 'foo': 100})
    assert result['_key'] == key
    assert (await ecol.get(key))['foo'] == 100

    # Test update edge with return_old and return_new set to True
    result = await ecol.update(
        {'_key': key, 'foo': 100},
        return_old=True,
        return_new=True
    )
    assert 'old' in result
    assert 'new' in result
    assert 'edge' in result
    old_rev = (await ecol.get(key))['_rev']

    # Test update edge with multiple field changes
    result = await ecol.update({'_key': key, 'foo': 200, 'bar': 300})
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] == 300
    old_rev = result['_rev']

    # Test update edge with correct revision
    result = await ecol.update({'_key': key, '_rev': old_rev, 'bar': 400})
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] == 400
    old_rev = result['_rev']

    if ecol.context != 'transaction':
        # Test update edge with bad revision
        new_rev = old_rev + '1'
        with assert_raises(DocumentRevisionError, DocumentUpdateError):
            await ecol.update({'_key': key, '_rev': new_rev, 'bar': 500})
        assert (await ecol.get(key))['foo'] == 200
        assert (await ecol.get(key))['bar'] == 400

    # Test update edge in missing edge collection
    with assert_raises(DocumentUpdateError) as err:
        await bad_ecol.update({'_key': key, 'bar': 500})
    assert err.value.error_code in {11, 1228}
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] == 400

    # Test update edge with sync option
    result = await ecol.update({'_key': key, 'bar': 500}, sync=True)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] == 500
    old_rev = result['_rev']

    # Test update edge with silent option
    assert await ecol.update({'_key': key, 'bar': 600}, silent=True) is True
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] == 600
    assert (await ecol.get(key))['_rev'] != old_rev
    old_rev = (await ecol.get(key))['_rev']

    # Test update edge without keep_none option
    result = await ecol.update({'_key': key, 'bar': None}, keep_none=True)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] is None
    old_rev = result['_rev']

    # Test update edge with keep_none option
    result = await ecol.update({'_key': key, 'foo': None}, keep_none=False)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert 'foo' not in await ecol.get(key)
    assert (await ecol.get(key))['bar'] is None

    # Test replace edge with a single field change
    edge['foo'] = 100
    result = await ecol.replace(edge)
    assert result['_key'] == key
    assert (await ecol.get(key))['foo'] == 100

    # Test replace edge with return_old and return_new set to True
    result = await ecol.replace(edge, return_old=True, return_new=True)
    assert 'old' in result
    assert 'new' in result
    assert 'edge' in result
    old_rev = (await ecol.get(key))['_rev']

    # Test replace edge with silent set to True
    edge['bar'] = 200
    assert await ecol.replace(edge, silent=True) is True
    assert (await ecol.get(key))['foo'] == 100
    assert (await ecol.get(key))['bar'] == 200
    assert (await ecol.get(key))['_rev'] != old_rev
    old_rev = (await ecol.get(key))['_rev']

    # Test replace edge with multiple field changes
    edge['foo'] = 200
    edge['bar'] = 300
    result = await ecol.replace(edge)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 200
    assert (await ecol.get(key))['bar'] == 300
    old_rev = result['_rev']

    # Test replace edge with correct revision
    edge['foo'] = 300
    edge['bar'] = 400
    edge['_rev'] = old_rev
    result = await ecol.replace(edge)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 300
    assert (await ecol.get(key))['bar'] == 400
    old_rev = result['_rev']

    edge['bar'] = 500
    if ecol.context != 'transaction':
        # Test replace edge with bad revision
        edge['_rev'] = old_rev + key
        with assert_raises(DocumentRevisionError, DocumentReplaceError) as err:
            await ecol.replace(edge)
        assert err.value.error_code in {1200, 1903}
        assert (await ecol.get(key))['foo'] == 300
        assert (await ecol.get(key))['bar'] == 400

    # Test replace edge with bad database
    with assert_raises(DocumentReplaceError) as err:
        await bad_ecol.replace(edge)
    assert err.value.error_code in {11, 1228}
    assert (await ecol.get(key))['foo'] == 300
    assert (await ecol.get(key))['bar'] == 400

    # Test replace edge with sync option
    result = await ecol.replace(edge, sync=True, check_rev=False)
    assert result['_key'] == key
    assert result['_old_rev'] == old_rev
    assert (await ecol.get(key))['foo'] == 300
    assert (await ecol.get(key))['bar'] == 500

    # Test delete edge with bad revision
    if ecol.context != 'transaction':
        old_rev = (await ecol.get(key))['_rev']
        edge['_rev'] = old_rev + '1'
        with assert_raises(DocumentRevisionError, DocumentDeleteError) as err:
            await ecol.delete(edge, check_rev=True)
        assert err.value.error_code in {1200, 1903}
        edge['_rev'] = old_rev
        assert await ecol.has(edge)

    # Test delete missing edge
    with assert_raises(DocumentDeleteError) as err:
        await ecol.delete(bad_document_key, ignore_missing=False)
    assert err.value.error_code == 1202
    if ecol.context != 'transaction':
        assert not await ecol.delete(bad_document_key, ignore_missing=True)

    # Test delete existing edge with sync set to True
    assert await ecol.delete(edge, sync=True, check_rev=False) is True
    if ecol.context != 'transaction':
        assert await ecol.get(edge) is None
    assert not await ecol.has(edge)

    # Test delete existing edge with return_old set to True
    await ecol.insert(edge)
    result = await ecol.delete(edge, return_old=True, check_rev=False)
    assert 'old' in result
    assert not await ecol.has(edge)
    await empty_collection(ecol)


async def test_vertex_edges(db, bad_db):
    graph_name = generate_graph_name()
    vcol_name = generate_col_name()
    ecol_name = generate_col_name()

    # Prepare test documents
    anna = {'_id': '{}/anna'.format(vcol_name)}
    dave = {'_id': '{}/dave'.format(vcol_name)}
    josh = {'_id': '{}/josh'.format(vcol_name)}
    mary = {'_id': '{}/mary'.format(vcol_name)}
    tony = {'_id': '{}/tony'.format(vcol_name)}

    # Create test graph, vertex and edge collections
    school = await db.create_graph(graph_name)

    vcol = await school.create_vertex_collection(vcol_name)
    ecol = await school.create_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[vcol_name],
        to_vertex_collections=[vcol_name]
    )
    # Insert test vertices into the graph
    await vcol.insert(anna)
    await vcol.insert(dave)
    await vcol.insert(josh)
    await vcol.insert(mary)
    await vcol.insert(tony)

    # Insert test edges into the graph
    await ecol.link(anna, dave)
    await ecol.link(josh, dave)
    await ecol.link(mary, dave)
    await ecol.link(tony, dave)
    await ecol.link(dave, anna)

    # Test edges with default direction (both)
    result = await ecol.edges(dave)
    assert 'stats' in result
    assert 'filtered' in result['stats']
    assert 'scanned_index' in result['stats']
    assert len(result['edges']) == 5

    result = await ecol.edges(anna)
    assert len(result['edges']) == 2

    # Test edges with direction set to "in"
    result = await ecol.edges(dave, direction='in')
    assert len(result['edges']) == 4

    result = await ecol.edges(anna, direction='in')
    assert len(result['edges']) == 1

    # Test edges with direction set to "out"
    result = await ecol.edges(dave, direction='out')
    assert len(result['edges']) == 1

    result = await ecol.edges(anna, direction='out')
    assert len(result['edges']) == 1

    bad_graph = bad_db.graph(graph_name)
    with assert_raises(EdgeListError) as err:
        await bad_graph.edge_collection(ecol_name).edges(dave)
    assert err.value.error_code in {11, 1228}


async def test_edge_management_via_graph(graph, ecol, fvcol, fvdocs, tvcol, tvdocs):
    for vertex in fvdocs:
        await fvcol.insert(vertex)
    for vertex in tvdocs:
        await tvcol.insert(vertex)
    await empty_collection(ecol)

    # Get a random "from" vertex
    from_vertex = await fvcol.random()
    assert await graph.has_vertex(from_vertex)

    # Get a random "to" vertex
    to_vertex = await tvcol.random()
    assert await graph.has_vertex(to_vertex)

    # Test insert edge via graph object
    result = await graph.insert_edge(
        ecol.name,
        {'_from': from_vertex['_id'], '_to': to_vertex['_id']}
    )
    assert await ecol.has(result['_key'])
    assert await graph.has_edge(result['_id'])
    assert await ecol.count() == 1

    # Test link vertices via graph object
    result = await graph.link(ecol.name, from_vertex, to_vertex)
    assert await ecol.has(result['_key'])
    assert await ecol.count() == 2
    edge_id = result['_id']

    # Test get edge via graph object
    assert (await graph.edge(edge_id))['_id'] == edge_id

    # Test list edges via graph object
    result = await graph.edges(ecol.name, from_vertex, direction='out')
    assert 'stats' in result
    assert len(result['edges']) == 2

    result = await graph.edges(ecol.name, from_vertex, direction='in')
    assert 'stats' in result
    assert len(result['edges']) == 0

    # Test update edge via graph object
    result = await graph.update_edge({'_id': edge_id, 'foo': 100})
    assert result['_id'] == edge_id
    assert (await ecol.get(edge_id))['foo'] == 100

    # Test replace edge via graph object
    result = await graph.replace_edge({
        '_id': edge_id,
        '_from': from_vertex['_id'],
        '_to': to_vertex['_id'],
        'bar': 200
    })
    assert result['_id'] == edge_id
    assert 'foo' not in await ecol.get(edge_id)
    assert (await ecol.get(edge_id))['bar'] == 200

    # Test delete edge via graph object
    assert await graph.delete_edge(edge_id) is True
    assert not await ecol.has(edge_id)
    assert await ecol.count() == 1


async def test_traverse(db):
    # Create test graph, vertex and edge collections
    school = await db.create_graph(generate_graph_name())
    profs = await school.create_vertex_collection(generate_col_name())
    classes = await school.create_vertex_collection(generate_col_name())
    teaches = await school.create_edge_definition(
        edge_collection=generate_col_name(),
        from_vertex_collections=[profs.name],
        to_vertex_collections=[classes.name]
    )
    # Insert test vertices into the graph
    await profs.insert({'_key': 'anna', 'name': 'Professor Anna'})
    await profs.insert({'_key': 'andy', 'name': 'Professor Andy'})
    await classes.insert({'_key': 'CSC101', 'name': 'Introduction to CS'})
    await classes.insert({'_key': 'MAT223', 'name': 'Linear Algebra'})
    await classes.insert({'_key': 'STA201', 'name': 'Statistics'})
    await classes.insert({'_key': 'MAT101', 'name': 'Calculus I'})
    await classes.insert({'_key': 'MAT102', 'name': 'Calculus II'})

    # Insert test edges into the graph
    await teaches.insert({
        '_from': '{}/anna'.format(profs.name),
        '_to': '{}/CSC101'.format(classes.name)
    })
    await teaches.insert({
        '_from': '{}/anna'.format(profs.name),
        '_to': '{}/STA201'.format(classes.name)
    })
    await teaches.insert({
        '_from': '{}/anna'.format(profs.name),
        '_to': '{}/MAT223'.format(classes.name)
    })
    await teaches.insert({
        '_from': '{}/andy'.format(profs.name),
        '_to': '{}/MAT101'.format(classes.name)
    })
    await teaches.insert({
        '_from': '{}/andy'.format(profs.name),
        '_to': '{}/MAT102'.format(classes.name)
    })
    await teaches.insert({
        '_from': '{}/andy'.format(profs.name),
        '_to': '{}/MAT223'.format(classes.name)
    })

    # Traverse the graph with default settings
    result = await school.traverse('{}/anna'.format(profs.name))
    visited = await extract('_key', result['vertices'])
    assert visited == ['CSC101', 'MAT223', 'STA201', 'anna']

    for path in result['paths']:
        for vertex in path['vertices']:
            assert set(vertex) == {'_id', '_key', '_rev', 'name'}
        for edge in path['edges']:
            assert set(edge) == {'_id', '_key', '_rev', '_to', '_from'}

    result = await school.traverse('{}/andy'.format(profs.name))
    visited = await extract('_key', result['vertices'])
    assert visited == ['MAT101', 'MAT102', 'MAT223', 'andy']

    # Traverse the graph with an invalid start vertex
    with assert_raises(GraphTraverseError):
        await school.traverse('invalid')

    with assert_raises(GraphTraverseError):
        bad_col_name = generate_col_name()
        await school.traverse('{}/hanna'.format(bad_col_name))

    with assert_raises(GraphTraverseError):
        await school.traverse('{}/anderson'.format(profs.name))

    # Travers the graph with max iteration of 0
    with assert_raises(GraphTraverseError):
        await school.traverse('{}/andy'.format(profs.name), max_iter=0)

    # Traverse the graph with max depth of 0
    result = await school.traverse('{}/andy'.format(profs.name), max_depth=0)
    assert await extract('_key', result['vertices']) == ['andy']

    result = await school.traverse('{}/anna'.format(profs.name), max_depth=0)
    assert await extract('_key', result['vertices']) == ['anna']

    # Traverse the graph with min depth of 2
    result = await school.traverse('{}/andy'.format(profs.name), min_depth=2)
    assert await extract('_key', result['vertices']) == []

    result = await school.traverse('{}/anna'.format(profs.name), min_depth=2)
    assert await extract('_key', result['vertices']) == []

    # Traverse the graph with DFS and BFS
    result = await school.traverse(
        {'_id': '{}/anna'.format(profs.name)},
        strategy='dfs',
        direction='any',
    )
    dfs_vertices = await extract('_key', result['vertices'])

    result = await school.traverse(
        {'_id': '{}/anna'.format(profs.name)},
        strategy='bfs',
        direction='any'
    )
    bfs_vertices = await extract('_key', result['vertices'])

    assert sorted(dfs_vertices) == sorted(bfs_vertices)

    # Traverse the graph with filter function
    result = await school.traverse(
        {'_id': '{}/andy'.format(profs.name)},
        filter_func='if (vertex._key == "MAT101") {return "exclude";} return;'
    )
    assert await extract('_key', result['vertices']) == ['MAT102', 'MAT223', 'andy']

    # Traverse the graph with global uniqueness (should be same as before)
    result = await school.traverse(
        {'_id': '{}/andy'.format(profs.name)},
        vertex_uniqueness='global',
        edge_uniqueness='global',
        filter_func='if (vertex._key == "MAT101") {return "exclude";} return;'
    )
    assert await extract('_key', result['vertices']) == ['MAT102', 'MAT223', 'andy']

    with assert_raises(DocumentParseError) as err:
        await school.traverse({})
    assert err.value.message == 'field "_id" required'



from __future__ import absolute_import, unicode_literals

from six import string_types
import pytest

from aioarangodb.collection import StandardCollection
from aioarangodb.exceptions import (
    CollectionChecksumError,
    CollectionConfigureError,
    CollectionLoadError,
    CollectionPropertiesError,
    CollectionRenameError,
    CollectionRevisionError,
    CollectionRotateJournalError,
    CollectionStatisticsError,
    CollectionTruncateError,
    CollectionUnloadError,
    CollectionCreateError,
    CollectionListError,
    CollectionDeleteError,
    CollectionRecalculateCountError
)
from aioarangodb.tests.helpers import assert_raises, extract, generate_col_name
pytestmark = pytest.mark.asyncio


async def test_collection_attributes(db, col, username):
    assert col.context in ['default', 'async', 'batch', 'transaction']
    assert col.username == username
    assert col.db_name == db.name
    assert col.name.startswith('test_collection')
    assert repr(col) == '<StandardCollection {}>'.format(col.name)


async def test_collection_misc_methods(col, bad_col, cluster):
    # Test get properties
    properties = await col.properties()
    assert properties['name'] == col.name
    assert properties['system'] is False

    # Test get properties with bad collection
    with assert_raises(CollectionPropertiesError) as err:
        await bad_col.properties()
    assert err.value.error_code in {11, 1228}

    # Test configure properties
    prev_sync = properties['sync']
    properties = await col.configure(
        sync=not prev_sync,
        journal_size=10000000
    )
    assert properties['name'] == col.name
    assert properties['system'] is False
    assert properties['sync'] is not prev_sync

    # Test configure properties with bad collection
    with assert_raises(CollectionConfigureError) as err:
        await bad_col.configure(sync=True, journal_size=10000000)
    assert err.value.error_code in {11, 1228}

    # Test get statistics
    stats = await col.statistics()
    assert isinstance(stats, dict)
    assert 'indexes' in stats

    # Test get statistics with bad collection
    with assert_raises(CollectionStatisticsError) as err:
        await bad_col.statistics()
    assert err.value.error_code in {11, 1228}

    # Test get revision
    assert isinstance(await col.revision(), string_types)

    # Test get revision with bad collection
    with assert_raises(CollectionRevisionError) as err:
        await bad_col.revision()
    assert err.value.error_code in {11, 1228}

    # Test load into memory
    assert await col.load() is True

    # Test load with bad collection
    with assert_raises(CollectionLoadError) as err:
        await bad_col.load()
    assert err.value.error_code in {11, 1228}

    # Test unload from memory
    assert await col.unload() is True

    # Test unload with bad collection
    with assert_raises(CollectionUnloadError) as err:
        await bad_col.unload()
    assert err.value.error_code in {11, 1228}

    # Test rotate journal
    try:
        assert isinstance(await col.rotate(), bool)
    except CollectionRotateJournalError as err:
        assert err.error_code in {404, 1105}

    # Test rotate journal with bad collection
    with assert_raises(CollectionRotateJournalError) as err:
        await bad_col.rotate()
    assert err.value.error_code in {11, 1228}

    if cluster:
        await col.insert({})
    else:
        # Test checksum with empty collection
        assert int(await col.checksum(with_rev=True, with_data=False)) == 0
        assert int(await col.checksum(with_rev=True, with_data=True)) == 0
        assert int(await col.checksum(with_rev=False, with_data=False)) == 0
        assert int(await col.checksum(with_rev=False, with_data=True)) == 0

        # Test checksum with non-empty collection
        await col.insert({})
        assert int(await col.checksum(with_rev=True, with_data=False)) > 0
        assert int(await col.checksum(with_rev=True, with_data=True)) > 0
        assert int(await col.checksum(with_rev=False, with_data=False)) > 0
        assert int(await col.checksum(with_rev=False, with_data=True)) > 0

        # Test checksum with bad collection
        with assert_raises(CollectionChecksumError) as err:
            await bad_col.checksum()
        assert err.value.error_code in {11, 1228}

    # Test preconditions
    assert await col.count() == 1

    # Test truncate collection
    assert await col.truncate() is True
    assert await col.count() == 0

    # Test truncate with bad collection
    with assert_raises(CollectionTruncateError) as err:
        await bad_col.truncate()
    assert err.value.error_code in {11, 1228}

    # Test recalculate count
    assert await col.recalculate_count() is True

    # Test recalculate count with bad collection
    with assert_raises(CollectionRecalculateCountError) as err:
        await bad_col.recalculate_count()
    assert err.value.error_code in {11, 1228}


async def test_collection_management(db, bad_db, cluster):
    # Test create collection
    col_name = generate_col_name()
    assert await db.has_collection(col_name) is False

    col = await db.create_collection(
        name=col_name,
        sync=True,
        compact=False,
        journal_size=7774208,
        system=False,
        volatile=False,
        key_generator='traditional',
        user_keys=False,
        key_increment=9,
        key_offset=100,
        edge=True,
        shard_count=2,
        shard_fields=['test_attr'],
        index_bucket_count=10,
        replication_factor=1,
        shard_like='',
        sync_replication=False,
        enforce_replication_factor=False,
        sharding_strategy='community-compat',
        smart_join_attribute='test',
        write_concern=1
    )
    assert await db.has_collection(col_name) is True

    properties = await col.properties()
    assert 'key_options' in properties
    assert properties['name'] == col_name
    assert properties['sync'] is True
    assert properties['system'] is False

    # Test create duplicate collection
    with assert_raises(CollectionCreateError) as err:
        await db.create_collection(col_name)
    assert err.value.error_code == 1207

    # Test list collections
    assert all(
        entry['name'].startswith('test_collection')
        or entry['name'].startswith('_')
        for entry in await db.collections()
    )

    # Test list collections with bad database
    with assert_raises(CollectionListError) as err:
        await bad_db.collections()
    assert err.value.error_code in {11, 1228}

    # Test get collection object
    test_col = db.collection(col.name)
    assert isinstance(test_col, StandardCollection)
    assert test_col.name == col.name

    test_col = db[col.name]
    assert isinstance(test_col, StandardCollection)
    assert test_col.name == col.name

    # Test delete collection
    assert await db.delete_collection(col_name, system=False) is True
    assert col_name not in await extract('name', await db.collections())

    # Test drop missing collection
    with assert_raises(CollectionDeleteError) as err:
        await db.delete_collection(col_name)
    assert err.value.error_code == 1203
    assert await db.delete_collection(col_name, ignore_missing=True) is False

    if not cluster:
        # Test rename collection
        new_name = generate_col_name()
        col = await db.create_collection(new_name)
        assert await col.rename(new_name) is True
        assert col.name == new_name
        assert repr(col) == '<StandardCollection {}>'.format(new_name)

        # Try again (the operation should be idempotent)
        assert await col.rename(new_name) is True
        assert col.name == new_name
        assert repr(col) == '<StandardCollection {}>'.format(new_name)

        # Test rename with bad collection
        with assert_raises(CollectionRenameError) as err:
            await bad_db.collection(new_name).rename(new_name)
        assert err.value.error_code in {11, 1228}



from setuptools import find_packages, setup

import re


def load_reqs(filename):
    with open(filename) as reqs_file:
        return [
            re.sub('==', '>=', line) for line in reqs_file.readlines()
            if not re.match('\s*#', line)
        ]


version = open('VERSION').read().rstrip('\n')
requirements = load_reqs('requirements.txt')
test_requirements = load_reqs('test-requirements.txt')

try:
    README = open('README.rst').read() + '\n\n' + open('CHANGELOG.rst').read()
except IOError:
    README = None

setup(
    name='aioarangodb',
    version=version,
    description='AsyncIO ArangoDB driver',
    long_description=README,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    url='http://github.com/bloodbare/aioarangodb',
    author='Ramon Navarro',
    author_email='ramon.nb@gmail.com',
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    install_requires=requirements,
    test_suite='tests',
    tests_require=test_requirements
)



from __future__ import absolute_import, unicode_literals

import asyncio

import pytest
from six import string_types

from aioarangodb.exceptions import (
    PregelJobCreateError,
    PregelJobGetError,
    PregelJobDeleteError
)
from aioarangodb.tests.helpers import (
    assert_raises,
    generate_string
)
pytestmark = pytest.mark.asyncio


async def test_pregel_attributes(db, username):
    assert db.pregel.context in ['default', 'async', 'batch', 'transaction']
    assert db.pregel.username == username
    assert db.pregel.db_name == db.name
    assert repr(db.pregel) == '<Pregel in {}>'.format(db.name)


async def test_pregel_management(db, graph, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    # Test create pregel job
    job_id = await db.pregel.create_job(
        graph.name,
        'pagerank',
        store=False,
        max_gss=100,
        thread_count=1,
        async_mode=False,
        result_field='result',
        algorithm_params={'threshold': 0.000001}
    )
    assert isinstance(job_id, int)

    # Test create pregel job with unsupported algorithm
    with assert_raises(PregelJobCreateError) as err:
        await db.pregel.create_job(graph.name, 'invalid')
    assert err.value.error_code in {4, 10, 1600}

    # Test get existing pregel job
    job = await db.pregel.job(job_id)
    assert isinstance(job['state'], string_types)
    assert isinstance(job['aggregators'], dict)
    assert isinstance(job['gss'], int)
    assert isinstance(job['received_count'], int)
    assert isinstance(job['send_count'], int)
    assert 'total_runtime' in job

    # Test delete existing pregel job
    assert await db.pregel.delete_job(job_id) is True
    await asyncio.sleep(0.2)
    with assert_raises(PregelJobGetError) as err:
        await db.pregel.job(job_id)
    assert err.value.error_code in {4, 10, 1600}

    # Test delete missing pregel job
    with assert_raises(PregelJobDeleteError) as err:
        await db.pregel.delete_job(generate_string())
    assert err.value.error_code in {4, 10, 1600}



import time

from aioarangodb.executor import (
    AsyncExecutor,
    BatchExecutor,
    TransactionExecutor,
)
from aioarangodb.job import BatchJob


class TestAsyncExecutor(AsyncExecutor):

    def __init__(self, connection):
        super(TestAsyncExecutor, self).__init__(
            connection=connection,
            return_result=True
        )

    async def execute(self, request, response_handler):
        job = await AsyncExecutor.execute(self, request, response_handler)
        while job.status() != 'done':
            time.sleep(.01)
        return job.result()


class TestBatchExecutor(BatchExecutor):

    def __init__(self, connection):
        super(TestBatchExecutor, self).__init__(
            connection=connection,
            return_result=True
        )

    async def execute(self, request, response_handler):
        self._committed = False
        self._queue.clear()

        job = BatchJob(response_handler)
        self._queue[job.id] = (request, job)
        await self.commit()
        return await job.result()


class TestTransactionExecutor(TransactionExecutor):

    # noinspection PyMissingConstructor
    def __init__(self, connection):
        self._conn = connection

    async def execute(self, request, response_handler):
        if request.read is request.write is request.exclusive is None:
            resp = self._conn.send_request(request)
            return response_handler(resp)

        super(TestTransactionExecutor, self).__init__(
            connection=self._conn,
            sync=True,
            allow_implicit=False,
            lock_timeout=0,
            read=request.read,
            write=request.write,
            exclusive=request.exclusive
        )
        result = await TransactionExecutor.execute(self, request, response_handler)
        await self.commit()
        return result



from __future__ import absolute_import, unicode_literals

import pytest

from aioarangodb.database import TransactionDatabase
from aioarangodb.exceptions import (
    TransactionExecuteError,
    TransactionInitError,
    TransactionStatusError,
    TransactionCommitError,
    TransactionAbortError
)
from aioarangodb.tests.helpers import extract
pytestmark = pytest.mark.asyncio


async def test_transaction_execute_raw(db, col, docs):
    # Test execute raw transaction
    doc = docs[0]
    key = doc['_key']
    result = await db.execute_transaction(
        command='''
        function (params) {{
            var db = require('internal').db;
            db.{col}.save({{'_key': params.key, 'val': 1}});
            return true;
        }}
        '''.format(col=col.name),
        params={'key': key},
        write=[col.name],
        read=[col.name],
        sync=False,
        timeout=1000,
        max_size=100000,
        allow_implicit=True,
        intermediate_commit_count=10,
        intermediate_commit_size=10000
    )
    assert result is True
    assert await col.has(doc) and (await col.get(key))['val'] == 1

    # Test execute invalid transaction
    with pytest.raises(TransactionExecuteError) as err:
        await db.execute_transaction(command='INVALID COMMAND')
    assert err.value.error_code == 10


async def test_transaction_init(db, bad_db, col, username):
    txn_db = await db.begin_transaction()

    assert isinstance(txn_db, TransactionDatabase)
    assert txn_db.username == username
    assert txn_db.context == 'transaction'
    assert txn_db.db_name == db.name
    assert txn_db.name == db.name
    assert txn_db.transaction_id is not None
    assert repr(txn_db) == '<TransactionDatabase {}>'.format(db.name)

    txn_col = txn_db.collection(col.name)
    assert txn_col.username == username
    assert txn_col.context == 'transaction'
    assert txn_col.db_name == db.name

    txn_aql = txn_db.aql
    assert txn_aql.username == username
    assert txn_aql.context == 'transaction'
    assert txn_aql.db_name == db.name

    with pytest.raises(TransactionInitError) as err:
        await bad_db.begin_transaction()
    assert err.value.error_code in {11, 1228}


async def test_transaction_status(db, col, docs):
    txn_db = await db.begin_transaction(read=col.name)
    assert await txn_db.transaction_status() == 'running'

    await txn_db.commit_transaction()
    assert await txn_db.transaction_status() == 'committed'

    txn_db = await db.begin_transaction(read=col.name)
    assert await txn_db.transaction_status() == 'running'

    await txn_db.abort_transaction()
    assert await txn_db.transaction_status() == 'aborted'

    # Test transaction_status with an illegal transaction ID
    txn_db._executor._id = 'illegal'
    with pytest.raises(TransactionStatusError) as err:
        await txn_db.transaction_status()
    assert err.value.error_code in {10, 1655}


async def test_transaction_commit(db, col, docs):
    txn_db = await db.begin_transaction(
        read=col.name,
        write=col.name,
        exclusive=[],
        sync=True,
        allow_implicit=False,
        lock_timeout=1000,
        max_size=10000
    )
    txn_col = txn_db.collection(col.name)

    assert '_rev' in await txn_col.insert(docs[0])
    assert '_rev' in await txn_col.delete(docs[0])
    assert '_rev' in await txn_col.insert(docs[1])
    assert '_rev' in await txn_col.delete(docs[1])
    assert '_rev' in await txn_col.insert(docs[2])
    await txn_db.commit_transaction()

    assert await extract('_key', await col.all()) == [docs[2]['_key']]
    assert await txn_db.transaction_status() == 'committed'

    # Test commit_transaction with an illegal transaction ID
    txn_db._executor._id = 'illegal'
    with pytest.raises(TransactionCommitError) as err:
        await txn_db.commit_transaction()
    assert err.value.error_code in {10, 1655}


async def test_transaction_abort(db, col, docs):
    txn_db = await db.begin_transaction(write=col.name)
    txn_col = txn_db.collection(col.name)

    assert '_rev' in await txn_col.insert(docs[0])
    assert '_rev' in await txn_col.delete(docs[0])
    assert '_rev' in await txn_col.insert(docs[1])
    assert '_rev' in await txn_col.delete(docs[1])
    assert '_rev' in await txn_col.insert(docs[2])
    await txn_db.abort_transaction()

    assert await extract('_key', await col.all()) == []
    assert await txn_db.transaction_status() == 'aborted'

    txn_db._executor._id = 'illegal'
    with pytest.raises(TransactionAbortError) as err:
        await txn_db.abort_transaction()
    assert err.value.error_code in {10, 1655}


async def test_transaction_graph(db, graph, fvcol, fvdocs):
    txn_db = await db.begin_transaction(write=fvcol.name)
    vcol = txn_db.graph(graph.name).vertex_collection(fvcol.name)

    await vcol.insert(fvdocs[0])
    assert await vcol.count() == 1

    await vcol.delete(fvdocs[0])
    assert await vcol.count() == 0

    await txn_db.commit()



from __future__ import absolute_import, unicode_literals
import pytest

from aioarangodb.exceptions import (
    ViewCreateError,
    ViewDeleteError,
    ViewGetError,
    ViewListError,
    ViewRenameError,
    ViewReplaceError,
    ViewUpdateError
)
from aioarangodb.tests.helpers import assert_raises, generate_view_name
pytestmark = pytest.mark.asyncio


async def test_view_management(db, bad_db, cluster):
    # Test create view
    view_name = generate_view_name()
    bad_view_name = generate_view_name()
    view_type = 'arangosearch'

    result = await db.create_view(
        view_name,
        view_type,
        {'consolidationIntervalMsec': 50000}
    )
    assert 'id' in result
    assert result['name'] == view_name
    assert result['type'] == view_type
    assert result['consolidation_interval_msec'] == 50000
    view_id = result['id']

    # Test create duplicate view
    with assert_raises(ViewCreateError) as err:
        await db.create_view(
            view_name,
            view_type,
            {'consolidationIntervalMsec': 50000}
        )
    assert err.value.error_code == 1207

    # Test list views
    result = await db.views()
    assert len(result) == 1
    view = result[0]
    assert view['id'] == view_id
    assert view['name'] == view_name
    assert view['type'] == view_type

    # Test list views with bad database
    with assert_raises(ViewListError) as err:
        await bad_db.views()
    assert err.value.error_code in {11, 1228}

    # Test get view
    view = await db.view(view_name)
    assert view['id'] == view_id
    assert view['name'] == view_name
    assert view['type'] == view_type
    assert view['consolidation_interval_msec'] == 50000

    # Test get missing view
    with assert_raises(ViewGetError) as err:
        await db.view(bad_view_name)
    assert err.value.error_code == 1203

    # Test update view
    view = await db.update_view(view_name, {'consolidationIntervalMsec': 70000})
    assert view['id'] == view_id
    assert view['name'] == view_name
    assert view['type'] == view_type
    assert view['consolidation_interval_msec'] == 70000

    # Test update view with bad database
    with assert_raises(ViewUpdateError) as err:
        await bad_db.update_view(view_name, {'consolidationIntervalMsec': 80000})
    assert err.value.error_code in {11, 1228}

    # Test replace view
    view = await db.replace_view(view_name, {'consolidationIntervalMsec': 40000})
    assert view['id'] == view_id
    assert view['name'] == view_name
    assert view['type'] == view_type
    assert view['consolidation_interval_msec'] == 40000

    # Test replace view with bad database
    with assert_raises(ViewReplaceError) as err:
        await bad_db.replace_view(view_name, {'consolidationIntervalMsec': 7000})
    assert err.value.error_code in {11, 1228}

    if cluster:
        new_view_name = view_name
    else:
        # Test rename view
        new_view_name = generate_view_name()
        assert await db.rename_view(view_name, new_view_name) is True
        result = await db.views()
        assert len(result) == 1
        view = result[0]
        assert view['id'] == view_id
        assert view['name'] == new_view_name

        # Test rename missing view
        with assert_raises(ViewRenameError) as err:
            await db.rename_view(bad_view_name, view_name)
        assert err.value.error_code == 1203

    # Test delete view
    assert await db.delete_view(new_view_name) is True
    assert len(await db.views()) == 0

    # Test delete missing view
    with assert_raises(ViewDeleteError) as err:
        await db.delete_view(new_view_name)
    assert err.value.error_code == 1203

    # Test delete missing view with ignore_missing set to True
    assert await db.delete_view(view_name, ignore_missing=True) is False


async def test_arangosearch_view_management(db, bad_db, cluster):
    # Test create arangosearch view
    view_name = generate_view_name()
    result = await db.create_arangosearch_view(
        view_name,
        {'consolidationIntervalMsec': 50000}
    )
    assert 'id' in result
    assert result['name'] == view_name
    assert result['type'].lower() == 'arangosearch'
    assert result['consolidation_interval_msec'] == 50000
    view_id = result['id']

    # Test create duplicate arangosearch view
    with assert_raises(ViewCreateError) as err:
        await db.create_arangosearch_view(
            view_name,
            {'consolidationIntervalMsec': 50000}
        )
    assert err.value.error_code == 1207

    result = await db.views()
    if not cluster:
        assert len(result) == 1
        view = result[0]
        assert view['id'] == view_id
        assert view['name'] == view_name
        assert view['type'] == 'arangosearch'

    # Test update arangosearch view
    view = await db.update_arangosearch_view(
        view_name,
        {'consolidationIntervalMsec': 70000}
    )
    assert view['id'] == view_id
    assert view['name'] == view_name
    assert view['type'].lower() == 'arangosearch'
    assert view['consolidation_interval_msec'] == 70000

    # Test update arangosearch view with bad database
    with assert_raises(ViewUpdateError) as err:
        await bad_db.update_arangosearch_view(
            view_name,
            {'consolidationIntervalMsec': 70000}
        )
    assert err.value.error_code in {11, 1228}

    # Test replace arangosearch view
    view = await db.replace_arangosearch_view(
        view_name,
        {'consolidationIntervalMsec': 40000}
    )
    assert view['id'] == view_id
    assert view['name'] == view_name
    assert view['type'] == 'arangosearch'
    assert view['consolidation_interval_msec'] == 40000

    # Test replace arangosearch with bad database
    with assert_raises(ViewReplaceError) as err:
        await bad_db.replace_arangosearch_view(
            view_name,
            {'consolidationIntervalMsec': 70000}
        )
    assert err.value.error_code in {11, 1228}

    # Test delete arangosearch view
    assert await db.delete_view(view_name, ignore_missing=False) is True



from __future__ import absolute_import, unicode_literals

import json

from six import string_types

__all__ = ['ArangoClient']

from .connection import (
    BasicConnection,
    JWTConnection,
    JWTSuperuserConnection
)
from .database import StandardDatabase
from .exceptions import ServerConnectionError
from .http import DefaultHTTPClient
from .resolver import (
    SingleHostResolver,
    RandomHostResolver,
    RoundRobinHostResolver
)
from .version import __version__


class ArangoClient(object):
    """ArangoDB client.

    :param hosts: Host URL or list of URLs (coordinators in a cluster).
    :type hosts: [str | unicode]
    :param host_resolver: Host resolver. This parameter used for clusters (when
        multiple host URLs are provided). Accepted values are "roundrobin" and
        "random". Any other value defaults to round robin.
    :type host_resolver: str | unicode
    :param http_client: User-defined HTTP client.
    :type http_client: arango.http.HTTPClient
    :param serializer: User-defined JSON serializer. Must be a callable
        which takes a JSON data type object as its only argument and return
        the serialized string. If not given, ``json.dumps`` is used by default.
    :type serializer: callable
    :param deserializer: User-defined JSON de-serializer. Must be a callable
        which takes a JSON serialized string as its only argument and return
        the de-serialized object. If not given, ``json.loads`` is used by
        default.
    :type deserializer: callable
    """

    def __init__(self,
                 hosts='http://127.0.0.1:8529',
                 host_resolver='roundrobin',
                 http_client=None,
                 serializer=json.dumps,
                 deserializer=json.loads):
        if isinstance(hosts, string_types):
            self._hosts = [host.strip('/') for host in hosts.split(',')]
        else:
            self._hosts = [host.strip('/') for host in hosts]

        host_count = len(self._hosts)
        if host_count == 1:
            self._host_resolver = SingleHostResolver()
        elif host_resolver == 'random':
            self._host_resolver = RandomHostResolver(host_count)
        else:
            self._host_resolver = RoundRobinHostResolver(host_count)

        self._http = http_client or DefaultHTTPClient()
        self._serializer = serializer
        self._deserializer = deserializer
        self._sessions = [self._http.create_session(h) for h in self._hosts]

    def __repr__(self):
        return '<ArangoClient {}>'.format(','.join(self._hosts))

    @property
    def hosts(self):
        """Return the list of ArangoDB host URLs.

        :return: List of ArangoDB host URLs.
        :rtype: [str | unicode]
        """
        return self._hosts

    @property
    def version(self):
        """Return the client version.

        :return: Client version.
        :rtype: str | unicode
        """
        return __version__

    async def close(self):
        for session in self._sessions:
            await session.close()

    async def db(
            self,
            name='_system',
            username='root',
            password='',
            verify=False,
            auth_method='basic',
            superuser_token=None):
        """Connect to an ArangoDB database and return the database API wrapper.

        :param name: Database name.
        :type name: str | unicode
        :param username: Username for basic authentication.
        :type username: str | unicode
        :param password: Password for basic authentication.
        :type password: str | unicode
        :param verify: Verify the connection by sending a test request.
        :type verify: bool
        :param auth_method: HTTP authentication method. Accepted values are
            "basic" (default) and "jwt". If set to "jwt", the token is
            refreshed automatically using ArangoDB username and password. This
            assumes that the clocks of the server and client are synchronized.
        :type auth_method: str | unicode
        :param superuser_token: User generated token for superuser access.
            If set, parameters **username**, **password** and **auth_method**
            are ignored. This token is not refreshed automatically.
        :type superuser_token: str | unicode
        :return: Standard database API wrapper.
        :rtype: arango.database.StandardDatabase
        :raise arango.exceptions.ServerConnectionError: If **verify** was set
            to True and the connection fails.
        """
        if superuser_token is not None:
            connection = JWTSuperuserConnection(
                hosts=self._hosts,
                host_resolver=self._host_resolver,
                sessions=self._sessions,
                db_name=name,
                http_client=self._http,
                serializer=self._serializer,
                deserializer=self._deserializer,
                superuser_token=superuser_token
            )
        elif auth_method == 'basic':
            connection = BasicConnection(
                hosts=self._hosts,
                host_resolver=self._host_resolver,
                sessions=self._sessions,
                db_name=name,
                username=username,
                password=password,
                http_client=self._http,
                serializer=self._serializer,
                deserializer=self._deserializer,
            )
        elif auth_method == 'jwt':
            connection = JWTConnection(
                hosts=self._hosts,
                host_resolver=self._host_resolver,
                sessions=self._sessions,
                db_name=name,
                username=username,
                password=password,
                http_client=self._http,
                serializer=self._serializer,
                deserializer=self._deserializer,
            )
            await connection.refresh_token()
        else:
            raise ValueError('invalid auth_method: {}'.format(auth_method))

        if verify:
            try:
                await connection.ping()
            except ServerConnectionError as err:
                raise err
            except Exception as err:
                raise ServerConnectionError('bad connection: {}'.format(err))

        return StandardDatabase(connection)



from __future__ import absolute_import, unicode_literals
import pytest
from datetime import datetime

from six import string_types

from aioarangodb.exceptions import (
    DatabaseCreateError,
    DatabaseDeleteError,
    DatabaseListError,
    DatabasePropertiesError,
    ServerDetailsError,
    ServerEchoError,
    ServerLogLevelError,
    ServerLogLevelSetError,
    ServerMetricsError,
    ServerReadLogError,
    ServerReloadRoutingError,
    ServerRequiredDBVersionError,
    ServerRoleError,
    ServerStatusError,
    ServerStatisticsError,
    ServerTimeError,
    ServerVersionError,
    ServerEngineError,
)
from aioarangodb.tests.helpers import assert_raises, generate_db_name
pytestmark = pytest.mark.asyncio


async def test_database_attributes(db, username):
    assert db.context in ['default', 'async', 'batch', 'transaction']
    assert db.username == username
    assert db.db_name == db.name
    assert db.name.startswith('test_database')
    assert db.conn is not None
    assert repr(db) == '<StandardDatabase {}>'.format(db.name)


async def test_database_misc_methods(sys_db, db, bad_db):
    # Test get properties
    properties = await db.properties()
    assert 'id' in properties
    assert 'path' in properties
    assert properties['name'] == db.name
    assert properties['system'] is False

    # Test get properties with bad database
    with assert_raises(DatabasePropertiesError) as err:
        await bad_db.properties()
    assert err.value.error_code in {11, 1228}

    # Test get server version
    assert isinstance(await db.version(), string_types)

    # Test get server version with bad database
    with assert_raises(ServerVersionError) as err:
        await bad_db.version()
    assert err.value.error_code in {11, 1228}

    # Test get server details
    details = await db.details()
    assert 'architecture' in details
    assert 'server-version' in details

    # Test get server details with bad database
    with assert_raises(ServerDetailsError) as err:
        await bad_db.details()
    assert err.value.error_code in {11, 1228}

    # Test get server required database version
    version = await db.required_db_version()
    assert isinstance(version, string_types)

    # Test get server target version with bad database
    with assert_raises(ServerRequiredDBVersionError):
        await bad_db.required_db_version()

    # Test get server metrics
    metrics = await db.metrics()
    assert isinstance(metrics, string_types)

    # Test get server statistics with bad database
    with assert_raises(ServerMetricsError) as err:
        await bad_db.metrics()
    assert err.value.error_code in {11, 1228}

    # Test get server statistics
    statistics = await db.statistics(description=False)
    assert isinstance(statistics, dict)
    assert 'time' in statistics
    assert 'system' in statistics
    assert 'server' in statistics

    # Test get server statistics with description
    description = await db.statistics(description=True)
    assert isinstance(description, dict)
    assert 'figures' in description
    assert 'groups' in description

    # Test get server statistics with bad database
    with assert_raises(ServerStatisticsError) as err:
        await bad_db.statistics()
    assert err.value.error_code in {11, 1228}

    # Test get server role
    assert await db.role() in {
        'SINGLE',
        'COORDINATOR',
        'PRIMARY',
        'SECONDARY',
        'UNDEFINED'
    }

    # Test get server role with bad database
    with assert_raises(ServerRoleError) as err:
        await bad_db.role()
    assert err.value.error_code in {11, 1228}

    # Test get server status
    status = await db.status()
    assert 'host' in status
    assert 'operation_mode' in status
    assert 'server_info' in status
    assert 'read_only' in status['server_info']
    assert 'write_ops_enabled' in status['server_info']
    assert 'version' in status

    # Test get status with bad database
    with assert_raises(ServerStatusError) as err:
        await bad_db.status()
    assert err.value.error_code in {11, 1228}

    # Test get server time
    assert isinstance(await db.time(), datetime)

    # Test get server time with bad database
    with assert_raises(ServerTimeError) as err:
        await bad_db.time()
    assert err.value.error_code in {11, 1228}

    # Test echo (get last request)
    last_request = await db.echo()
    assert 'protocol' in last_request
    assert 'user' in last_request
    assert 'requestType' in last_request
    assert 'rawRequestBody' in last_request

    # Test echo with bad database
    with assert_raises(ServerEchoError) as err:
        await bad_db.echo()
    assert err.value.error_code in {11, 1228}

    # Test read_log with default parameters
    log = await db.read_log(upto='fatal')
    assert 'lid' in log
    assert 'level' in log
    assert 'text' in log
    assert 'total_amount' in log

    # Test read_log with specific parameters
    log = await db.read_log(
        level='error',
        start=0,
        size=100000,
        offset=0,
        search='test',
        sort='desc',
    )
    assert 'lid' in log
    assert 'level' in log
    assert 'text' in log
    assert 'total_amount' in log

    # Test read_log with bad database
    with assert_raises(ServerReadLogError) as err:
        await bad_db.read_log()
    assert err.value.error_code in {11, 1228}

    # Test reload routing
    assert isinstance(await db.reload_routing(), bool)

    # Test reload routing with bad database
    with assert_raises(ServerReloadRoutingError) as err:
        await bad_db.reload_routing()
    assert err.value.error_code in {11, 1228}

    # Test get log levels
    assert isinstance(await db.log_levels(), dict)

    # Test get log levels with bad database
    with assert_raises(ServerLogLevelError) as err:
        await bad_db.log_levels()
    assert err.value.error_code in {11, 1228}

    # Test set log levels
    new_levels = {
        'agency': 'DEBUG',
        'collector': 'INFO',
        'threads': 'WARNING'
    }
    result = await db.set_log_levels(**new_levels)
    for key, value in new_levels.items():
        assert result[key] == value
    for key, value in (await db.log_levels()).items():
        assert result[key] == value

    # Test set log levels with bad database
    with assert_raises(ServerLogLevelSetError):
        await bad_db.set_log_levels(**new_levels)

    # Test get storage engine
    engine = await db.engine()
    assert engine['name'] in ['mmfiles', 'rocksdb']
    assert 'supports' in engine

    # Test get storage engine with bad database
    with assert_raises(ServerEngineError) as err:
        await bad_db.engine()
    assert err.value.error_code in {11, 1228}


async def test_database_management(db, sys_db, bad_db):
    # Test list databases
    result = await sys_db.databases()
    assert '_system' in result

    # Test list databases with bad database
    with assert_raises(DatabaseListError):
        await bad_db.databases()

    # Test create database
    db_name = generate_db_name()
    assert await sys_db.has_database(db_name) is False
    assert await sys_db.create_database(
        name=db_name,
        replication_factor=1,
        write_concern=1,
        sharding="single"
    ) is True
    assert await sys_db.has_database(db_name) is True

    # Test create duplicate database
    with assert_raises(DatabaseCreateError) as err:
        await sys_db.create_database(db_name)
    assert err.value.error_code == 1207

    # Test create database without permissions
    with assert_raises(DatabaseCreateError) as err:
        await db.create_database(db_name)
    assert err.value.error_code == 1230

    # Test delete database without permissions
    with assert_raises(DatabaseDeleteError) as err:
        await db.delete_database(db_name)
    assert err.value.error_code == 1230

    # Test delete database
    assert await sys_db.delete_database(db_name) is True
    assert db_name not in await sys_db.databases()

    # Test delete missing database
    with assert_raises(DatabaseDeleteError) as err:
        await sys_db.delete_database(db_name)
    assert err.value.error_code in {11, 1228}
    assert await sys_db.delete_database(db_name, ignore_missing=True) is False



from __future__ import absolute_import, unicode_literals

__all__ = ['Request']


class Request(object):
    """HTTP request.

    :param method: HTTP method in lowercase (e.g. "post").
    :type method: str | unicode
    :param endpoint: API endpoint.
    :type endpoint: str | unicode
    :param headers: Request headers.
    :type headers: dict
    :param params: URL parameters.
    :type params: dict
    :param data: Request payload.
    :type data: str | unicode | bool | int | list | dict |
        aiohttp.MultipartWriter
    :param read: Names of collections read during transaction.
    :type read: str | unicode | [str | unicode]
    :param write: Name(s) of collections written to during transaction with
        shared access.
    :type write: str | unicode | [str | unicode]
    :param exclusive: Name(s) of collections written to during transaction
        with exclusive access.
    :type exclusive: str | unicode | [str | unicode]
    :param deserialize: Whether the response body can be deserialized.
    :type deserialize: bool
    :ivar method: HTTP method in lowercase (e.g. "post").
    :vartype method: str | unicode
    :ivar endpoint: API endpoint.
    :vartype endpoint: str | unicode
    :ivar headers: Request headers.
    :vartype headers: dict
    :ivar params: URL (query) parameters.
    :vartype params: dict
    :ivar data: Request payload.
    :vartype data: str | unicode | bool | int | list | dict
    :ivar read: Names of collections read during transaction.
    :vartype read: str | unicode | [str | unicode] | None
    :ivar write: Name(s) of collections written to during transaction with
        shared access.
    :vartype write: str | unicode | [str | unicode] | None
    :ivar exclusive: Name(s) of collections written to during transaction
        with exclusive access.
    :vartype exclusive: str | unicode | [str | unicode] | None
    :ivar deserialize: Whether the response body can be deserialized.
    :vartype deserialize: bool
    """

    __slots__ = (
        'method',
        'endpoint',
        'headers',
        'params',
        'data',
        'read',
        'write',
        'exclusive',
        'deserialize',
        'files'
    )

    def __init__(self,
                 method,
                 endpoint,
                 headers=None,
                 params=None,
                 data=None,
                 read=None,
                 write=None,
                 exclusive=None,
                 deserialize=True):
        self.method = method
        self.endpoint = endpoint
        self.headers = {
            'content-type': 'application/json',
            'charset': 'utf-8'
        }
        if headers is not None:
            for field in headers:
                self.headers[field.lower()] = headers[field]

        # Sanitize URL params.
        if params is not None:
            for key, val in params.items():
                if isinstance(val, bool):
                    params[key] = int(val)

        self.params = params
        self.data = data
        self.read = read
        self.write = write
        self.exclusive = exclusive
        self.deserialize = deserialize
